#
# Copyright (C) 2019 Luca Pasqualini
# University of Siena - Artificial Intelligence Laboratory - SAILab
#
#
# USienaRL is licensed under a BSD 3-Clause.
#
# You should have received a copy of the license along with this
# work. If not, see <https://opensource.org/licenses/BSD-3-Clause>.

# Import packages

import logging
import random

# Import required src

from usienarl import ExplorationPolicy, Interface, SpaceType


class EpsilonGreedyExplorationPolicy(ExplorationPolicy):
    """
    Epsilon greedy exploration policy.

    According to this policy, the best action predicted by the agent inner model is used if a certain value,
    the exploration rate, is below a certain threshold.

    When updating the policy, the exploration rate decreases by its decay value.

    Supported action spaces:
        - discrete
        - continuous
    """

    def __init__(self,
                 exploration_rate_max: float, exploration_rate_min: float,
                 exploration_rate_decay: float):
        # Define epsilon-greedy exploration policy attributes
        self._exploration_rate_max: float = exploration_rate_max
        self._exploration_rate_min: float = exploration_rate_min
        self._exploration_rate_decay: float = exploration_rate_decay
        # Define epsilon-greedy empty exploration policy attributes
        self._exploration_rate: float = None
        # Generate the base exploration policy
        super().__init__()
        # Define the types of allowed action space types
        self._supported_action_space_types.append(SpaceType.discrete)
        self._supported_action_space_types.append(SpaceType.continuous)

    def _define(self):
        pass

    def initialize(self,
                   logger: logging.Logger,
                   session):
        # Reset exploration rate to its starting value (the max)
        self._exploration_rate = self._exploration_rate_max

    def act(self,
            logger: logging.Logger,
            session,
            interface: Interface,
            all_actions, best_action):
        # Choose an action according to the epsilon greedy approach: best action or random action
        if self._exploration_rate > 0 and random.uniform(0, 1.0) < self._exploration_rate:
            action = interface.get_random_agent_action(logger, session)
        else:
            action = best_action
        return action

    def update(self,
               logger: logging.Logger,
               session):
        # Decrease the exploration rate by its decay value
        self._exploration_rate = max(self._exploration_rate_min, self._exploration_rate - self._exploration_rate_decay)



#
# Copyright (C) 2019 Luca Pasqualini
# University of Siena - Artificial Intelligence Laboratory - SAILab
#
#
# USienaRL is licensed under a BSD 3-Clause.
#
# You should have received a copy of the license along with this
# work. If not, see <https://opensource.org/licenses/BSD-3-Clause>.

# Import packages

import logging
import numpy

# Import required src

from usienarl import ExplorationPolicy, Interface, SpaceType


class BoltzmannExplorationPolicy(ExplorationPolicy):
    """
    Boltzmann exploration policy.

    According to this policy, a sampling of the actions predicted by the agent inner model is executed to decide the
    best next action. The sampling is done using a softmax over the actions values with an associated temperature.

    When updating the policy, the temperature decreases by its decay value.

    Supported action spaces:
        - discrete
    """

    def __init__(self,
                 temperature_max: float, temperature_min: float,
                 temperature_decay: float):
        # Define boltzmann exploration policy attributes
        self._temperature_max: float = temperature_max
        self._temperature_min: float = temperature_min
        self._temperature_decay: float = temperature_decay
        # Define epsilon-greedy empty exploration policy attributes
        self._temperature: float = None
        # Generate the base explorer
        super().__init__()
        # Define the types of allowed action space types
        self._supported_action_space_types.append(SpaceType.discrete)

    def _define(self):
        pass

    def initialize(self, logger: logging.Logger, session):
        # Reset temperature to its starting value (the max)
        self._temperature = self._temperature_max

    def act(self,
            logger: logging.Logger,
            session,
            interface: Interface,
            all_actions, best_action):
        # Act according to boltzmann approach: get the softmax over all the actions predicted by the model
        output = self._softmax(all_actions / self._temperature)
        # Get a random action value (random output) using the softmax as probability distribution
        action_value = numpy.random.choice(output[0], p=output[0])
        # Return the chose action as the index of such chosen action value
        return numpy.argmax(output[0] == action_value)

    def update(self,
               logger: logging.Logger,
               session):
        # Decrease the exploration rate by its decay value
        self._temperature = max(self._temperature_min, self._temperature - self._temperature_decay)

    @staticmethod
    def _softmax(array):
        """
        Compute the softmax of an array.

        :param array: the array of which to compute the softmax
        :return: the softmax value
        """
        # Make sure the length of the shape of the given array is 2
        assert len(array.shape) == 2
        # Get the element with max value in the given array as the normalization factor
        normalization_factor = numpy.max(array, axis=1)
        # Increase the shape size of the normalization factor to allow broadcasting
        normalization_factor = normalization_factor[:, numpy.newaxis]
        # Apply the normalization
        numerator = numpy.exp(array - normalization_factor)
        # Compute the denominator by summing all the normalized numerator elements
        denominator = numpy.sum(numerator, axis=1)
        # Increase the shape size of the denominator to allow broadcasting
        denominator = denominator[:, numpy.newaxis]
        # Return the softmax result
        return numerator / denominator



#
# Copyright (C) 2019 Luca Pasqualini
# University of Siena - Artificial Intelligence Laboratory - SAILab
#
#
# USienaRL is licensed under a BSD 3-Clause.
#
# You should have received a copy of the license along with this
# work. If not, see <https://opensource.org/licenses/BSD-3-Clause>.

# Import packages

import logging

# Import required src

from usienarl import Interface, SpaceType


class ExplorationPolicy:
    """
    Base abstract class for exploration policies.
    An exploration policy act similar to a model, predicting actions and being updated if necessary. It also requires
    to be generated before being used.

    For each exploration policy, a set of supported action space types is defined. To define a novel exploration policy,
    implement the abstract class in a new specific child class.
    """

    def __init__(self):
        # Define empty exploration policy attributes
        self._agent_action_space_type: SpaceType = None
        self._agent_action_space_shape = None
        self._supported_action_space_types: [] = []

    def generate(self,
                 logger: logging.Logger,
                 agent_action_space_type: SpaceType, agent_action_space_shape) -> bool:
        """
        Generate the exploration policy. Checks if the exploration policy is compatible with the given action space and
        define the exploration policy itself.

        :param logger: the logger used to print the exploration policy information, warnings and errors
        :param agent_action_space_type: the space type of the agent action space
        :param agent_action_space_shape: the space shape of the agent action space
        :return: True if generation is successful, False otherwise
        """
        logger.info("Generating exploration policy...")
        # Set action space type and shape
        self._agent_action_space_type = agent_action_space_type
        self._agent_action_space_shape = agent_action_space_shape
        # Check if action space type is supported
        action_space_type_supported: bool = False
        for space_type in self._supported_action_space_types:
            if self._agent_action_space_type == space_type:
                action_space_type_supported = True
                break
        if not action_space_type_supported:
            logger.error("Error during setup of exploration policy: action space type not supported")
            return False
        # Define the exploration policy
        self._define()
        # Return success
        return True

    def _define(self):
        """
        Define the exploration policy. This is always called during generation.
        """
        raise NotImplementedError()

    def initialize(self,
                   logger: logging.Logger,
                   session):
        """
        Reset the exploration policy to its starting state, usually resetting the exploration rate to its max value.

        :param logger: the logger used to print the exploration policy information, warnings and errors
        :param session: the session of tensorflow currently running, if any
        """
        raise NotImplementedError()

    def act(self,
            logger: logging.Logger,
            session,
            interface: Interface,
            all_actions, best_action):
        """
        Act according to the model actions and/or predicted action.

        :param logger: the logger used to print the exploration policy information, warnings and errors
        :param session: the session of tensorflow currently running, if any
        :param interface: the interface between the agent using the exploration policy and the environment
        :param all_actions: all the action values predicted by the model
        :param best_action: the best (usually highest) action value predicted by the model
        """
        # Abstract method, definition should be implemented on a child class basis
        raise NotImplementedError()

    def update(self,
               logger: logging.Logger,
               session):
        """
        Update the exploration policy, usually changing the exploration rate.

        :param logger: the logger used to print the exploration policy information, warnings and errors
        :param session: the session of tensorflow currently running, if any
        """
        raise NotImplementedError()



#
# Copyright (C) 2019 Luca Pasqualini
# University of Siena - Artificial Intelligence Laboratory - SAILab
#
#
# USienaRL is licensed under a BSD 3-Clause.
#
# You should have received a copy of the license along with this
# work. If not, see <https://opensource.org/licenses/BSD-3-Clause>.

# Import scripts

from .config import Config, LayerType
from .environment import Environment, SpaceType
from .interface import Interface
from .agent import Agent
from .model import Model
from .experiment import Experiment
from .exploration_policy import ExplorationPolicy

# Import functions

from .functions import command_line_parse, run_experiment, watch_experiment




#
# Copyright (C) 2019 Luca Pasqualini
# University of Siena - Artificial Intelligence Laboratory - SAILab
#
#
# USienaRL is licensed under a BSD 3-Clause.
#
# You should have received a copy of the license along with this
# work. If not, see <https://opensource.org/licenses/BSD-3-Clause>.

# Import scripts

from .boltzmann import BoltzmannExplorationPolicy
from .epsilon_greedy import EpsilonGreedyExplorationPolicy



#
# Copyright (C) 2019 Luca Pasqualini
# University of Siena - Artificial Intelligence Laboratory - SAILab
#
#
# USienaRL is licensed under a BSD 3-Clause.
#
# You should have received a copy of the license along with this
# work. If not, see <https://opensource.org/licenses/BSD-3-Clause>.

# Import packages

import os.path
import datetime
import sys
import logging
import numpy
import argparse

# Import required src

from usienarl import Experiment


def _positive_int(value: str) -> int:
    """
    Positive integer check/conversion rule for arg-parse.

    :param value: the value to check if it's integer
    :return: the int value if positive
    """
    int_value: int = int(value)
    if int_value <= 0:
        raise argparse.ArgumentTypeError("%s is an invalid positive int value" % value)
    return int_value


def command_line_parse(watch: bool = False):
    """
    Parse command line using arg-parse and get the default experiment user data.

    :param watch: optional boolean flag defining if the command line parse is for a watch script or for a regular experiment script (the default one)
    :return: the parsed arguments: workspace directory, experiment iterations, CUDA devices, optional render during training flag, optional render during validation flag, optional render during testing flag
    """
    # Parse depending on the boolean watch flag
    parser = argparse.ArgumentParser()
    if watch:
        parser.add_argument("checkpoint_directory", type=str)
        parser.add_argument("iteration_number", type=int)
        parser.add_argument("CUDA_devices", type=str)
        parser.add_argument("-render", action="store_true")
        args: dict = vars(parser.parse_args())
        return args["checkpoint_directory"], args["iteration_number"], args["CUDA_devices"], args["render"]
    parser.add_argument("workspace_directory", type=str)
    parser.add_argument("experiment_iterations", type=_positive_int)
    parser.add_argument("CUDA_devices", type=str)
    parser.add_argument("-render_during_training", action="store_true")
    parser.add_argument("-render_during_validation", action="store_true")
    parser.add_argument("-render_during_testing", action="store_true")
    args: dict = vars(parser.parse_args())
    return args["workspace_directory"], args["experiment_iterations"], args["CUDA_devices"], args["render_during_training"], args["render_during_validation"], args["render_during_testing"]


def run_experiment(experiment: Experiment,
                   training_episodes: int,
                   max_training_episodes: int, episode_length_max: int,
                   validation_episodes: int,
                   testing_episodes: int, test_cycles: int,
                   render_during_training: bool, render_during_validation: bool, render_during_test: bool,
                   workspace_path: str, file_name: str, logger: logging.Logger,
                   checkpoint_path: str = None, experiment_iterations_number: int = 1) -> []:
    """
    Run the given experiment with the given parameters. It automatically creates all the directories to store the
    experiment results, summaries and meta-graphs for each iteration.

    :param experiment: the experiment to run
    :param training_episodes: the number of training episodes per volley
    :param max_training_episodes: the maximum number of training episodes in total
    :param episode_length_max: the maximum number of steps for each episode
    :param validation_episodes: the number of validation episodes per volley
    :param testing_episodes: the number of test episodes per test cycle
    :param test_cycles: the number of test cycles
    :param render_during_training: flag to render or not the environment during training
    :param render_during_validation: flag to render or not the environment during validation
    :param render_during_test: flag to render or not the environment during test
    :param workspace_path: path to workspace directory
    :param file_name: the name of the experiment script
    :param logger: the logger used to record information, warnings and errors
    :param checkpoint_path: the optional checkpoint path to gather the model from, useful to further train an already trained model
    :param experiment_iterations_number: the number of iterations of the experiment to run, by default just one
    :return the list of saved metagraph paths of each iteration of the experiment
    """
    # Generate the workspace directory if not defined
    workspace_directory = os.path.dirname(workspace_path)
    if not os.path.isdir(workspace_directory):
        try:
            os.makedirs(workspace_directory)
        except FileExistsError:
            pass
    # Generate the file directory with date if not defined
    file_path: str = workspace_path + "/" + os.path.splitext(os.path.basename(file_name))[0] + "_" + datetime.date.today().strftime("%Y-%m-%d")
    if not os.path.isdir(file_path):
        try:
            os.makedirs(file_path)
        except FileExistsError:
            pass
    # Generate a directory for the current experiment if not defined
    experiment_path: str = file_path + "/" + experiment.name
    if not os.path.isdir(experiment_path):
        try:
            os.makedirs(experiment_path)
        except FileExistsError:
            pass
    # Generate a directory for each iteration of the experiment if not defined
    # Prepare a table in which to store all results and a list of metagraph paths
    experiment_results_table: dict = dict()
    experiment_metagraph_paths: [] = []
    for iteration in range(experiment_iterations_number):
        iteration_path: str = experiment_path
        # Set the iteration number to -1 if there is only one iteration
        if experiment_iterations_number <= 1:
            iteration = -1
        if iteration > -1:
            iteration_path = experiment_path + "/" + str(iteration)
            if not os.path.isdir(iteration_path):
                try:
                    os.makedirs(iteration_path)
                except FileExistsError:
                    pass
        # Generate a metagraph and a summary directory for each iteration of experiment if not defined
        metagraph_path: str = iteration_path + "/metagraph"
        if not os.path.isdir(metagraph_path):
            try:
                os.makedirs(metagraph_path)
            except FileExistsError:
                pass
        summary_path: str = iteration_path + "/summary"
        if not os.path.isdir(summary_path):
            try:
                os.makedirs(summary_path)
            except FileExistsError:
                pass
        # Setup of the logger for the current experiment
        # Reset logger handlers for current experiment
        logger.handlers = []
        # Generate a console and a file handler for the logger
        console_handler: logging.StreamHandler = logging.StreamHandler(sys.stdout)
        file_handler: logging.FileHandler = logging.FileHandler(iteration_path + "/info.log", "w+")
        # Set handlers properties
        console_handler.setLevel(logging.DEBUG)
        file_handler.setLevel(logging.DEBUG)
        formatter: logging.Formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
        console_handler.setFormatter(formatter)
        file_handler.setFormatter(formatter)
        # Add the handlers to the logger
        logger.addHandler(console_handler)
        logger.addHandler(file_handler)
        # Conduct the experiment
        if experiment.setup(summary_path, metagraph_path, logger, iteration):
            average_total_reward, max_total_reward, average_scaled_reward, max_scaled_reward, trained_episodes, success, metagraph_save_path = experiment.conduct(training_episodes, validation_episodes,
                                                                                                                                                                  max_training_episodes, episode_length_max,
                                                                                                                                                                  testing_episodes, test_cycles,
                                                                                                                                                                  logger,
                                                                                                                                                                  render_during_training,
                                                                                                                                                                  render_during_validation,
                                                                                                                                                                  render_during_test,
                                                                                                                                                                  checkpoint_path)
            # Save the result of the iteration in the table
            experiment_results_table[iteration] = (average_total_reward, max_total_reward, average_scaled_reward, max_scaled_reward, trained_episodes, success)
            # Save the metagraph path in the list if the experiment iteration is successful
            if success:
                experiment_metagraph_paths.append(metagraph_save_path)
    # Print the results in both a results log file and in the console
    # Reset logger handlers for current experiment
    logger.handlers = []
    results_console_handler: logging.StreamHandler = logging.StreamHandler(sys.stdout)
    results_file_handler: logging.FileHandler = logging.FileHandler(experiment_path + "/results.log", "w+")
    results_console_handler.setLevel(logging.INFO)
    results_file_handler.setLevel(logging.INFO)
    results_console_formatter: logging.Formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
    results_file_formatter: logging.Formatter = logging.Formatter("%(message)s")
    results_console_handler.setFormatter(results_console_formatter)
    results_file_handler.setFormatter(results_file_formatter)
    # Add the file handler to the logger and print table of results
    logger.addHandler(results_file_handler)
    logger.info(experiment.name + " results:\n")
    logger.info("experiment iteration \t\t average total reward \t\t\t average scaled reward \t\t\t max total reward \t\t\t max scaled reward \t\t\t trained_episodes \t\t\t success")
    for experiment_iteration, result in experiment_results_table.items():
        logger.info(str(experiment_iteration) + "\t\t\t\t" + str("%.3f" % result[0]) + "\t\t\t\t" + str("%.3f" % result[1]) + "\t\t\t\t" + str("%.3f" % result[2]) + "\t\t\t\t" + str("%.3f" % result[3]) + "\t\t\t" + str(result[4]) + "\t\t\t" + "YES" if result[5] else "NO")
    logger.info("\n")
    # Add also the console handler and print verbose results
    logger.addHandler(results_console_handler)
    # Group all the final scores, best scores and training episode counts
    average_total_reward_list: [] = []
    average_scaled_reward_list: [] = []
    max_total_reward_list: [] = []
    max_scaled_reward_list: [] = []
    trained_episodes_list: [] = []
    success_list: [] = []
    for experiment_iteration, result in experiment_results_table.items():
        average_total_reward_list.append(result[0])
        max_total_reward_list.append(result[1])
        average_scaled_reward_list.append(result[2])
        max_scaled_reward_list.append(result[3])
        trained_episodes_list.append(result[4])
        success_list.append(result[5])
    max_average_total_reward_index: int = numpy.argmax(average_total_reward_list)
    max_max_total_reward_index: int = numpy.argmax(max_total_reward_list)
    max_average_scaled_reward_index: int = numpy.argmax(average_scaled_reward_list)
    max_max_scaled_reward_index: int = numpy.argmax(max_scaled_reward_list)
    min_training_episodes_index: int = numpy.argmin(trained_episodes_list)
    # Print best results for each indicator
    logger.info("Maximum average total reward over all " + str(test_cycles * testing_episodes) + " test episodes: " + str("%.3f" % numpy.max(average_total_reward_list)) + " at experiment number " + str(max_average_total_reward_index))
    logger.info("Maximum max total score over " + str(testing_episodes) + " test episodes: " + str("%.3f" % numpy.max(max_total_reward_list)) + " at experiment number " + str(max_max_total_reward_index))
    logger.info("Maximum average scaled reward over all " + str(test_cycles * testing_episodes) + " test episodes: " + str("%.3f" % numpy.max(average_scaled_reward_list)) + " at experiment number " + str(max_average_scaled_reward_index))
    logger.info("Maximum max scaled score over " + str(testing_episodes) + " test episodes: " + str("%.3f" % numpy.max(max_scaled_reward_list)) + " at experiment number " + str(max_max_scaled_reward_index))
    logger.info("Minimum training episodes for validation: " + str(numpy.min(trained_episodes_list)) + " at experiment number " + str(min_training_episodes_index))
    # Print average results for each indicator
    average_average_total_reward: float = numpy.average(average_total_reward_list)
    logger.info("Average of average total reward over all experiments " + str("%.3f" % average_average_total_reward))
    average_max_total_reward: float = numpy.average(max_total_reward_list)
    logger.info("Average of max total reward over all experiments: " + str("%.3f" % average_max_total_reward))
    average_average_scaled_reward: float = numpy.average(average_scaled_reward_list)
    logger.info("Average of average scaled reward over all experiments " + str("%.3f" % average_average_scaled_reward))
    average_max_scaled_reward: float = numpy.average(max_scaled_reward_list)
    logger.info("Average of max scaled reward over all experiments: " + str("%.3f" % average_max_scaled_reward))
    average_training_episodes: float = numpy.average(trained_episodes_list)
    logger.info("Average of training episodes over all experiments: " + str("%.3f" % average_training_episodes))
    success_percentage: float = numpy.round(100.0 * numpy.count_nonzero(success_list) / len(success_list), 2)
    logger.info("Success percentage over all experiments: " + str("%.2f" % success_percentage))
    # Print standard deviation for each indicator
    # Compute a list of averages for array difference
    average_total_reward_average_list = numpy.full(len(average_total_reward_list), average_average_total_reward)
    max_total_reward_average_list = numpy.full(len(max_total_reward_list), average_max_total_reward)
    average_scaled_reward_average_list = numpy.full(len(average_scaled_reward_list), average_average_scaled_reward)
    max_scaled_reward_average_list = numpy.full(len(max_scaled_reward_list), average_max_scaled_reward)
    training_episodes_average_list = numpy.full(len(trained_episodes_list), average_training_episodes)
    # Compute the actual standard deviation for each indicator
    stdv_average_total_reward: float = numpy.sqrt(numpy.sum(numpy.power(average_total_reward_list - average_total_reward_average_list, 2)) / len(average_total_reward_list))
    logger.info("Standard deviation of average total reward over all experiments: " + str("%.3f" % stdv_average_total_reward) + ", which is a " + str("%.2f" % ((100 * stdv_average_total_reward) / average_average_total_reward)) + "% of the average")
    stdv_max_total_reward: float = numpy.sqrt(numpy.sum(numpy.power(max_total_reward_list - max_total_reward_average_list, 2)) / len(max_total_reward_list))
    logger.info("Standard deviation of max total reward over all experiments " + str("%.3f" % stdv_max_total_reward) + ", which is a " + str("%.2f" % ((100 * stdv_max_total_reward) / average_max_total_reward)) + "% of the average")
    stdv_average_scaled_reward: float = numpy.sqrt(numpy.sum(numpy.power(average_scaled_reward_list - average_scaled_reward_average_list, 2)) / len(average_scaled_reward_list))
    logger.info("Standard deviation of average scaled reward over all experiments: " + str("%.3f" % stdv_average_scaled_reward) + ", which is a " + str("%.2f" % ((100 * stdv_average_scaled_reward) / average_average_scaled_reward)) + "% of the average")
    stdv_max_scaled_reward: float = numpy.sqrt(numpy.sum(numpy.power(max_scaled_reward_list - max_scaled_reward_average_list, 2)) / len(max_scaled_reward_list))
    logger.info("Standard deviation of max scaled reward over all experiments " + str("%.3f" % stdv_max_scaled_reward) + ", which is a " + str("%.2f" % ((100 * stdv_max_scaled_reward) / average_max_scaled_reward)) + "% of the average")
    stdv_training_episodes: float = numpy.sqrt(numpy.sum(numpy.power(trained_episodes_list - training_episodes_average_list, 2)) / len(trained_episodes_list))
    logger.info("Standard deviation of training episodes over all experiments: " + str("%.3f" % stdv_training_episodes) + ", which is a " + str("%.2f" % ((100 * stdv_training_episodes) / average_training_episodes)) + "% of the average")
    # Return the list of saved metagraph of successful iterations of the experiment, if any, otherwise an empty list
    return experiment_metagraph_paths


def watch_experiment(experiment: Experiment,
                     testing_episodes: int, test_cycles: int,
                     episode_length_max: int,
                     render: bool,
                     logger: logging.Logger,
                     checkpoint_path: str):
    """
    Run the given experiment with the given parameters. It automatically creates all the directories to store the
    experiment results, summaries and meta-graphs for each iteration.

    :param experiment: the experiment to run
    :param testing_episodes: the number of test episodes per test cycle
    :param test_cycles: the number of test cycles
    :param episode_length_max: the maximum number of steps for each episode
    :param render: flag to render or not the environment
    :param logger: the logger used to record information, warnings and errors
    :param checkpoint_path: the checkpoint path to gather the model from
    """
    # Setup of the logger for the current experiment
    # Reset logger handlers for current experiment
    logger.handlers = []
    # Generate a console handler for the logger
    console_handler: logging.StreamHandler = logging.StreamHandler(sys.stdout)
    # Set handler properties
    console_handler.setLevel(logging.DEBUG)
    formatter: logging.Formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
    console_handler.setFormatter(formatter)
    # Add the handler to the logger
    logger.addHandler(console_handler)
    # Watch the experiment
    if experiment.setup(None, None, logger):
        experiment.watch(episode_length_max,
                         testing_episodes, test_cycles,
                         logger, checkpoint_path,
                         render)



#
# Copyright (C) 2019 Luca Pasqualini
# University of Siena - Artificial Intelligence Laboratory - SAILab
#
#
# USienaRL is licensed under a BSD 3-Clause.
#
# You should have received a copy of the license along with this
# work. If not, see <https://opensource.org/licenses/BSD-3-Clause>.

# Import packages

import tensorflow
import numpy
import logging

# Import required src

from usienarl import Environment, Agent, Interface
from usienarl.interfaces import PassThroughInterface


class Experiment:
    """
    Base experiment abstract class.
    An experiment is a set of commands which allows to train or test or even both a certain agent on a certain
    environment using a certain interface to translate between the two.

    If an interface is not supplied, a default simple pass-through interface is used.

    To conduct an experiment, first make sure the setup phase is passed. Once an experiment is setup, the agent and
    the environment can be used.
    A default experiment consists of:
        - warm-up phase, where the agent acts in warm-up mode and which can be used or not depending on the agent
        inner model
        - volley phase, which consists in multiple sets of episodes dedicated to training after which the agent
        performance is validated
        - test phase, which consists in multiple cycles of test where the agent acts in inference mode and in which
        the agent performance is tested and which occurs only after validation is passed

    Each volley consist of two additional sub-phases:
        - train phase: where the agent acts in train mode and its internal model is updated
        - validation phase: where the agent acts in inference mode and its performance is tested

    To define your own experiment, implement the abstract class in a specific child class. A child class only requires
    to have implemented the methods used to validate and test the agent's model and conclude the experiment volley
    phase or the experiment as a whole.

    Two types of scores are produced, using the rewards given only by the environment:
        - total reward: the sum of all the rewards at each step in the episode, also called total episode reward
        - scaled reward: the average of all the rewards at each step in the episode, also called average step reward
    """

    def __init__(self,
                 name: str,
                 environment: Environment,
                 agent: Agent,
                 interface: Interface = None):
        # Define experiment attributes
        self.name: str = name
        self._environment: Environment = environment
        self._agent: Agent = agent
        self._interface: Interface = interface if interface is not None else PassThroughInterface(self._environment)
        # Define empty experiment attributes
        self._agent_saver = None
        self._metagraph_path: str = None
        self._tensorflow_config = None
        self._current_id: str = None
        self._trained_steps: int = None
        self._trained_episodes: int = None

    def setup(self,
              summary_path: str, metagraph_path: str,
              logger: logging.Logger,
              iteration: int = -1) -> bool:
        """
        Setup the experiment, preparing all of its component to execution. This must be called before conduct method.

        :param summary_path: the string path of the tensorboard summary directory to save during model training, set to None if the model is not being trained
        :param metagraph_path: the string path of the metagraph agent directory to save at the end of each train volley, set to None if the model is not being trained
        :param logger: the logger used to print the experiment information, warnings and errors
        :param iteration: number to append to the experiment name in all scopes and print statements (if not less than zero)
        :return: a boolean equals to True if the setup of the experiment is successful, False otherwise
        """
        # Save current experiment id using name and iteration
        self._current_id = self.name + "_" + str(iteration) if iteration > -1 else self.name
        # Start setup process
        logger.info("Setup of experiment " + self._current_id + "...")
        # Reset the tensorflow graph to a blank new graph
        tensorflow.reset_default_graph()
        # Execute the setup on the environment and on the agent and return if not both of them setup correctly
        logger.info("Environment setup...")
        if not self._environment.setup(logger):
            logger.info("Environment setup failed. Cannot setup the experiment!")
            return False
        logger.info("Environment setup is successful!")
        logger.info("Agent setup...")
        if not self._agent.setup(logger,
                                 self._interface.observation_space_type, self._interface.observation_space_shape,
                                 self._interface.agent_action_space_type, self._interface.agent_action_space_shape,
                                 self._current_id, summary_path):
            logger.info("Agent setup failed. Cannot setup the experiment!")
            return False
        logger.info("Agent setup is successful!")
        # If setup of both environment and agent is successful, reset internal experiment variables
        self._metagraph_path = metagraph_path
        self._trained_steps: int = 0
        self._trained_episodes: int = 0
        # Initialize the agent internal model saver
        self._agent_saver = tensorflow.train.Saver(self._agent.trainable_variables)
        if self._metagraph_path is not None:
            logger.info("Agent internal model will be saved after each train volley")
            logger.info("Agent internal model metagraph save path: " + self._metagraph_path)
        # Initialize tensorflow gpu configuration
        self._tensorflow_config = tensorflow.ConfigProto()
        self._tensorflow_config.gpu_options.allow_growth = True
        logger.info("Tensorflow configuration is successful!")
        # Setup is successful
        logger.info("Experiment setup is successful. Now can conduct the experiment")
        return True

    def _warmup(self,
                logger: logging.Logger,
                episodes: int, episode_length: int,
                session,
                render: bool = False):
        """
        Execute a volley of warm-up of the agent on the environment.

        :param logger: the logger used to print the experiment information, warnings and errors
        :param episodes: the number of episodes in which to run the agent in warm-up mode
        :param session: the session of tensorflow currently running
        :param render: boolean parameter deciding whether or not to render during warm-up
        """
        for episode in range(episodes):
            # Start the episode
            episode_rewards: [] = []
            state_current = self._environment.reset(logger, session)
            # Execute actions until the episode is completed or the maximum length is exceeded
            for step in range(episode_length):
                # Get the action decided by the agent with train policy
                observation_current = self._interface.environment_state_to_observation(logger, session, state_current)
                agent_action = self._agent.act_warmup(logger, session, self._interface, observation_current)
                # Get the next state with relative reward and completion flag
                environment_action = self._interface.agent_action_to_environment_action(logger, session, agent_action)
                state_next, reward, episode_done = self._environment.step(logger, environment_action, session)
                # Complete the step and send back information to the agent
                observation_next = self._interface.environment_state_to_observation(logger, session, state_next)
                # Set the observation next to None if final state is reached
                if episode_done:
                    observation_next = None
                self._agent.complete_step_warmup(logger, session, self._interface,
                                                 observation_current, agent_action, reward, observation_next,
                                                 step, episode, episodes)
                # Render if required
                if render:
                    self._environment.render(logger, session)
                # Add the reward to the list of rewards for this episode
                episode_rewards.append(reward)
                # Update the current state with the previously next state
                state_current = state_next
                # Check if the episode is completed
                if episode_done:
                    break
            # Compute total and average reward over the steps in the episode
            episode_total_reward: float = numpy.sum(numpy.array(episode_rewards))
            # Complete the episode and send back information to the agent
            self._agent.complete_episode_warmup(logger, session, self._interface,
                                                episode_rewards[-1], episode_total_reward,
                                                episode, episodes)

    def _train(self,
               logger: logging.Logger,
               episodes: int, episode_length: int, episodes_max: int,
               session,
               render: bool = False):
        """
        Execute a volley of training of the agent on the environment.

        :param logger: the logger used to print the experiment information, warnings and errors
        :param episodes: the number of episodes in which to run the agent in train mode
        :param session: the session of tensorflow currently running
        :param render: boolean parameter deciding whether or not to render during training
        :return: the float average of the total reward and of the average reward obtained in the played episodes
        """
        # Define arrays of volley rewards
        volley_average_rewards: numpy.ndarray = numpy.zeros(episodes, dtype=float)
        volley_total_rewards: numpy.ndarray = numpy.zeros(episodes, dtype=float)
        for episode in range(episodes):
            # Start the episode
            episode_rewards: [] = []
            state_current = self._environment.reset(logger, session)
            # Execute actions until the episode is completed or the maximum length is exceeded
            for step in range(episode_length):
                # Get the action decided by the agent with train policy
                observation_current = self._interface.environment_state_to_observation(logger, session, state_current)
                agent_action = self._agent.act_train(logger, session, self._interface, observation_current)
                # Get the next state with relative reward and completion flag
                environment_action = self._interface.agent_action_to_environment_action(logger, session, agent_action)
                state_next, reward, episode_done = self._environment.step(logger, environment_action, session)
                # Complete the step and send back information to the agent
                observation_next = self._interface.environment_state_to_observation(logger, session, state_next)
                # Set the observation next to None if final state is reached
                if episode_done:
                    observation_next = None
                self._agent.complete_step_train(logger, session, self._interface,
                                                observation_current, agent_action, reward, observation_next,
                                                step, self._trained_steps,
                                                episode, self._trained_episodes,
                                                episodes, episodes_max)
                # Render if required
                if render:
                    self._environment.render(logger, session)
                # Add the reward to the list of rewards for this episode
                episode_rewards.append(reward)
                # Update the current state with the previously next state
                state_current = state_next
                # Increase the number of trained steps
                self._trained_steps += 1
                # Check if the episode is completed
                if episode_done:
                    break
            # Compute total and average reward over the steps in the episode
            episode_total_reward: float = numpy.sum(numpy.array(episode_rewards))
            episode_average_reward: float = numpy.average(numpy.array(episode_rewards))
            # Complete the episode and send back information to the agent
            self._agent.complete_episode_train(logger, session, self._interface,
                                               episode_rewards[-1], episode_total_reward,
                                               self._trained_steps,
                                               episode, self._trained_episodes,
                                               episodes, episodes_max)
            # Increase the counter of trained episodes
            self._trained_episodes += 1
            # Add the episode rewards to the volley
            volley_total_rewards[episode] = episode_total_reward
            volley_average_rewards[episode] = episode_average_reward
        # Return the average of the total and of the averages over the episodes
        return numpy.average(volley_total_rewards), numpy.average(volley_average_rewards)

    def _inference(self,
                   logger: logging.Logger,
                   episodes: int, episode_length: int,
                   session,
                   render: bool = False):
        """
        Execute a volley of inference of the agent on the environment.

        :param logger: the logger used to print the experiment information, warnings and errors
        :param episodes: the number of episodes in which to run the agent in inference mode
        :param session: the session of tensorflow currently running
        :param render: boolean parameter deciding whether or not to render during inference
        :return: the float average of the total reward and of the average reward obtained in the played episodes
        """
        # Define arrays of volley rewards
        volley_average_rewards: numpy.ndarray = numpy.zeros(episodes, dtype=float)
        volley_total_rewards: numpy.ndarray = numpy.zeros(episodes, dtype=float)
        for episode in range(episodes):
            # Start the episode
            episode_rewards: [] = []
            state_current = self._environment.reset(logger, session)
            # Execute actions until the episode is completed or the maximum length is exceeded
            for step in range(episode_length):
                # Get the action decided by the agent with train policy
                observation_current = self._interface.environment_state_to_observation(logger, session, state_current)
                agent_action = self._agent.act_inference(logger, session, self._interface, observation_current)
                # Get the next state with relative reward and completion flag
                environment_action = self._interface.agent_action_to_environment_action(logger, session, agent_action)
                state_next, reward, episode_done = self._environment.step(logger, environment_action, session)
                # Complete the step and send back information to the agent
                observation_next = self._interface.environment_state_to_observation(logger, session, state_next)
                # Set the observation next to None if final state is reached
                if episode_done:
                    observation_next = None
                self._agent.complete_step_inference(logger, session, self._interface,
                                                    observation_current, agent_action, reward, observation_next,
                                                    step, episode, episodes)
                # Render if required
                if render:
                    self._environment.render(logger, session)
                # Add the reward to the list of rewards for this episode
                episode_rewards.append(reward)
                # Update the current state with the previously next state
                state_current = state_next
                # Check if the episode is completed
                if episode_done:
                    break
            # Compute total and average reward over the steps in the episode
            episode_total_reward: float = numpy.sum(numpy.array(episode_rewards))
            episode_average_reward: float = numpy.average(numpy.array(episode_rewards))
            # Complete the episode and send back information to the agent
            self._agent.complete_episode_inference(logger, session, self._interface,
                                                   episode_rewards[-1], episode_total_reward,
                                                   episode, episodes)
            # Add the episode rewards to the volley
            volley_total_rewards[episode] = episode_total_reward
            volley_average_rewards[episode] = episode_average_reward
        # Return the average of the total and of the averages over the episodes
        return numpy.average(volley_total_rewards), numpy.average(volley_average_rewards)

    def conduct(self,
                training_episodes_per_volley: int, validation_episodes_per_volley: int,
                training_episodes_max: int, episode_length_max: int,
                test_episodes_per_cycle: int, test_cycles: int,
                logger: logging.Logger,
                render_during_training: bool = False, render_during_validation: bool = False, render_during_test: bool = False,
                checkpoint_path: str = None):
        """
        Conduct the experiment of the given number of training (up to the given maximum) and validation episodes
        per volley, with the associated given number of test episodes per cycle and the given number of test cycles after
        the validation has passed.

        A volley of training is always considered to stop the training exactly when the agent performs well enough
        to validate, by checking the appropriate condition. When validation is successful, appropriate condition is also
        checked on the test cycles to decide if the experiment is successful or not.

        Conducting an experiment will generate a tensorflow session for that experiment. It is required that the experiment
        is ready to be conducted by checking the result of the setup method before attempting at conducting it.

        If an already trained model checkpoint is given, the last checkpoint in the given path folder is used to further
        train the model. The scope of the experiment and of the angent has to be the same across both models.

        :param training_episodes_per_volley: the number of training episodes per volley before trying to validate the agent
        :param validation_episodes_per_volley: the number of validation episodes per volley after the training in such interval
        :param training_episodes_max: the maximum number of training episodes allowed at all
        :param episode_length_max: the max length in steps of each episode
        :param test_episodes_per_cycle: the number of episodes to play for each test cycle after validation has passed
        :param test_cycles: the number of test cycles to execute after validation has passed
        :param logger: the logger used to print the experiment information, warnings and errors
        :param render_during_training: boolean flag to render the environment during training, default to False
        :param render_during_validation: boolean flag to render the environment during validation, default to False
        :param render_during_test: boolean flag to render the environment during test, default to False
        :param checkpoint_path: the optional checkpoint path to load the pre-trained model from, it should be a valid model, default to None
        :return: the average and the max of the total average reward, the average and the max of the average average reward with respect to all test cycles, the training episodes required to validate, a success flag and the location in which the saved metagraph is stored
        """
        logger.info("Conducting experiment " + self._current_id + "...")
        # Define the session
        with tensorflow.Session(config=self._tensorflow_config) as session:
            # Initialize the environment and the agent
            self._environment.initialize(logger, session)
            self._agent.initialize(logger, session)
            # Load the pre-trained model if required
            if checkpoint_path is not None:
                checkpoint = tensorflow.train.get_checkpoint_state(checkpoint_path)
                # If checkpoint exists restore from checkpoint
                if checkpoint and checkpoint.model_checkpoint_path:
                    self._agent_saver.restore(session, tensorflow.train.latest_checkpoint(checkpoint_path))
                    logger.info("Model graph stored at " + checkpoint_path + " loaded successfully!")
                else:
                    logger.error("Checkpoint path specified is wrong: no model can be accessed at " + checkpoint_path)
            # Execute pre-training if the agent requires pre-training
            if self._agent.warmup_episodes > 0:
                logger.info("Warming-up for " + str(self._agent.warmup_episodes) + " episodes...")
                self._warmup(logger,
                             self._agent.warmup_episodes,
                             episode_length_max,
                             session)
            # Save "last" volley training and validation rewards
            training_total_reward: float = 0.0
            training_average_reward: float = 0.0
            validation_total_reward: float = 0.0
            validation_average_reward: float = 0.0
            # Execute training until max training episodes number is reached or the validation score is above the threshold
            while self._trained_episodes < training_episodes_max:
                # Run train for training episodes per volley and get the average score
                logger.info("Training for " + str(training_episodes_per_volley) + " episodes...")
                training_total_reward, training_average_reward = self._train(logger,
                                                                             training_episodes_per_volley,
                                                                             episode_length_max,
                                                                             training_episodes_max,
                                                                             session,
                                                                             render_during_training)
                # Print training results
                logger.info("Training of " + str(training_episodes_per_volley) + " finished with following result:")
                logger.info("Average total reward over " + str(training_episodes_per_volley) + " training episodes after " + str(self._trained_episodes) + " total training episodes: " + str(training_total_reward))
                logger.info("Average scaled reward over " + str(training_episodes_per_volley) + " training episodes after " + str(self._trained_episodes) + " total training episodes: " + str(training_average_reward))
                logger.info("Total training steps up to now: " + str(self._trained_steps))
                # Save the agent internal _model at the current step
                logger.info("Saving the agent...")
                self._agent_saver.save(session, self._metagraph_path + "/" + self._current_id)
                # Run inference for validation episodes per volley and get the average score
                logger.info("Validating for " + str(validation_episodes_per_volley) + " episodes...")
                validation_total_reward, validation_average_reward = self._inference(logger,
                                                                                     validation_episodes_per_volley,
                                                                                     episode_length_max,
                                                                                     session,
                                                                                     render_during_validation)
                # Print validation results
                logger.info("Training of " + str(validation_episodes_per_volley) + " finished with following result:")
                logger.info("Average total reward over " + str(validation_episodes_per_volley) + " validation episodes after " + str(self._trained_episodes) + " total training episodes: " + str(validation_total_reward))
                logger.info("Average scaled reward over " + str(validation_episodes_per_volley) + " validation episodes after " + str(self._trained_episodes) + " total training episodes: " + str(validation_average_reward))
                # Check for validation
                if self._is_validated(validation_total_reward, validation_average_reward, training_total_reward, training_average_reward):
                    logger.info("Validation of the agent is successful")
                    break
            # Test the _model and get all cycles total and average rewards
            test_total_rewards: numpy.ndarray = numpy.zeros(test_cycles, dtype=float)
            test_average_rewards: numpy.ndarray = numpy.zeros(test_cycles, dtype=float)
            for test in range(test_cycles):
                # Run inference for test episodes per cycles and get the average score
                logger.info("Testing for " + str(test_episodes_per_cycle) + " episodes...")
                cycle_total_reward, cycle_average_reward = self._inference(logger,
                                                                           test_episodes_per_cycle,
                                                                           episode_length_max,
                                                                           session,
                                                                           render_during_test)
                # Print validation results
                logger.info("Testing of " + str(test_episodes_per_cycle) + " finished with following result:")
                logger.info("Average total reward over " + str(test_episodes_per_cycle) + " test episodes: " + str(cycle_total_reward))
                logger.info("Average scaled reward over " + str(test_episodes_per_cycle) + " test episodes: " + str(cycle_average_reward))
                # Save the rewards
                test_total_rewards[test] = cycle_total_reward
                test_average_rewards[test] = cycle_average_reward
            # Get the average and the best total and average rewards over all cycles
            average_test_total_reward: float = numpy.average(test_total_rewards)
            max_test_total_reward: float = numpy.max(test_total_rewards)
            average_test_average_reward: float = numpy.average(test_average_rewards)
            max_test_average_reward: float = numpy.max(test_average_rewards)
            # Print final results and outcome of the experiment
            logger.info("Average test total reward is " + str(average_test_total_reward) + " with " + str(self._trained_episodes) + " training episodes")
            logger.info("Average test scaled reward is " + str(average_test_average_reward) + " with " + str(self._trained_episodes) + " training episodes")
            logger.info("Best test total reward over " + str(test_episodes_per_cycle) + " and " + str(test_cycles) + " cycles is: " + str(max_test_total_reward))
            logger.info("Best test scaled reward over " + str(test_episodes_per_cycle) + " and " + str(test_cycles) + " cycles is: " + str(max_test_average_reward))
            # Check if the experiment is successful
            success: bool = self._is_successful(average_test_total_reward, average_test_average_reward, max_test_total_reward, max_test_average_reward, validation_total_reward, validation_average_reward, training_total_reward, training_average_reward)
            if success:
                logger.info("The experiment terminated successfully")
            else:
                logger.info("The experiment terminated without meeting the requirements")
            return average_test_total_reward, max_test_total_reward, average_test_average_reward, max_test_average_reward, self._trained_episodes, success, self._metagraph_path

    def watch(self,
              episode_length_max: int,
              test_episodes_per_cycle: int, test_cycles: int,
              logger: logging.Logger,
              checkpoint_path: str,
              render: bool = False):
        """
        Watch an experiment with an already trained agent given by the checkpoint path. The experiment needs to be
        setup and also trained somehow to access a proper checkpoint.

        Use this method to see how a saved agent model performs.

        :param episode_length_max: the max length in steps of each episode
        :param test_episodes_per_cycle: the number of episodes to play for each test cycle
        :param test_cycles: the number of test cycles to execute
        :param logger: the logger used to print the experiment information, warnings and errors
        :param checkpoint_path: the optional checkpoint path to load the pre-trained model from, it should be a valid model
        :param render: boolean flag to render the environment, default to True
        """
        logger.info("Watching experiment " + self._current_id + "...")
        # Define the session
        with tensorflow.Session(config=self._tensorflow_config) as session:
            # Initialize the environment and the agent
            self._environment.initialize(logger, session)
            self._agent.initialize(logger, session)
            # Load the pre-trained model if required
            if checkpoint_path is None:
                logger.error("Need to specify a checkpoint path to watch the experiment!")
                return
            else:
                checkpoint = tensorflow.train.get_checkpoint_state(checkpoint_path)
                # If checkpoint exists restore from checkpoint
                if checkpoint and checkpoint.model_checkpoint_path:
                    self._agent_saver.restore(session, tensorflow.train.latest_checkpoint(checkpoint_path))
                    logger.info("Model graph stored at " + checkpoint_path + " loaded successfully!")
                else:
                    logger.error("Checkpoint path specified is wrong: no model can be accessed at " + checkpoint_path)
                    return
            # Test the _model and get all cycles total and average rewards
            test_total_rewards: numpy.ndarray = numpy.zeros(test_cycles, dtype=float)
            test_average_rewards: numpy.ndarray = numpy.zeros(test_cycles, dtype=float)
            for test in range(test_cycles):
                # Run inference for test episodes per cycles and get the average score
                logger.info("Testing for " + str(test_episodes_per_cycle) + " episodes...")
                cycle_total_reward, cycle_average_reward = self._inference(logger,
                                                                           test_episodes_per_cycle,
                                                                           episode_length_max,
                                                                           session,
                                                                           render)
                # Print validation results
                logger.info("Testing of " + str(test_episodes_per_cycle) + " finished with following result:")
                logger.info("Average total reward over " + str(test_episodes_per_cycle) + " test episodes: " + str(cycle_total_reward))
                logger.info("Average scaled reward over " + str(test_episodes_per_cycle) + " test episodes: " + str(cycle_average_reward))
                # Save the rewards
                test_total_rewards[test] = cycle_total_reward
                test_average_rewards[test] = cycle_average_reward
            # Get the average and the best total and average rewards over all cycles
            average_test_total_reward: float = numpy.average(test_total_rewards)
            max_test_total_reward: float = numpy.max(test_total_rewards)
            average_test_average_reward: float = numpy.average(test_average_rewards)
            max_test_average_reward: float = numpy.max(test_average_rewards)
            # Print final results and outcome of the experiment
            logger.info("Average test total reward is " + str(average_test_total_reward) + " with " + str(self._trained_episodes) + " training episodes")
            logger.info("Average test scaled reward is " + str(average_test_average_reward) + " with " + str(self._trained_episodes) + " training episodes")
            logger.info("Best test total reward over " + str(test_episodes_per_cycle) + " and " + str(test_cycles) + " cycles is: " + str(max_test_total_reward))
            logger.info("Best test scaled reward over " + str(test_episodes_per_cycle) + " and " + str(test_cycles) + " cycles is: " + str(max_test_average_reward))

    def _is_validated(self,
                      average_validation_total_reward: float, average_validation_average_reward: float,
                      average_training_total_reward: float, average_training_average_reward: float) -> bool:
        """
        Check if the experiment is to be considered validated using the given parameters.

        :param average_validation_total_reward: the average total reward during last volley validation phase
        :param average_validation_average_reward: the average scaled reward during last volley validation phase
        :param average_training_total_reward: the average total reward during last volley training phase
        :param average_training_average_reward: the average scaled reward during last volley training phase
        :return: a boolean flag True if condition are satisfied, False otherwise
        """
        # Abstract method, definition should be implemented on a child class basis
        raise NotImplementedError()

    def _is_successful(self,
                       average_test_total_reward: float, average_test_average_reward: float,
                       max_test_total_reward: float, max_test_average_reward: float,
                       average_validation_total_reward: float, average_validation_average_reward: float,
                       average_training_total_reward: float, average_training_average_reward: float) -> bool:
        """
        Check if the experiment is to be considered successful using the given parameters.

        :param average_test_total_reward: the average total reward during all the test cycles
        :param average_test_average_reward: the average scaled reward during all the test cycles
        :param max_test_total_reward: the maximum total reward during all the test cycles
        :param max_test_average_reward: the maximum scaled reward during all the test cycles
        :param average_validation_total_reward: the average total reward during last volley validation phase
        :param average_validation_average_reward: the average scaled reward during last volley validation phase
        :param average_training_total_reward: the average total reward during last volley training phase
        :param average_training_average_reward: the average scaled reward during last volley training phase
        :return: a boolean flag True if condition are satisfied, False otherwise
        """
        # Abstract method, definition should be implemented on a child class basis
        raise NotImplementedError()



# Import packages

import tensorflow
import numpy
import scipy.signal

# Import required src

from usienarl import SpaceType, Config, Model


class Buffer:
    """
    A buffer for storing trajectories experienced by a VPG agent interacting with the environment,
    and using Generalized Advantage Estimation (GAE-Lambda) for calculating the advantages of state-action pairs.

    The buffer is dynamically resizable.

    The buffer contains list of states (or observations), actions (used as targets), values (computed by the value stream
    of the _model itself during prediction), advantages (computed by the buffer using GAE when a trajectory finishes and
    fed back up in the policy stream to drive the loss), rewards (used to compute rewards-to-go) and rewards-to-go
    (computed inside the buffer itself and used as weight for the targets action when training the value stream).
    """

    def __init__(self,
                 discount_factor: float, lambda_parameter: float):
        # Define buffer components
        self._states: [] = []
        self._actions: [] = []
        self._advantages: [] = []
        self._rewards: [] = []
        self._rewards_to_go: [] = []
        self._values: [] = []
        # Define parameters
        self._discount_factor: float = discount_factor
        self._lambda_parameter: float = lambda_parameter
        # Define buffer pointer
        self._pointer: int = 0
        self._path_start_index: int = 0

    def store(self,
              state, action, reward: float, value: float):
        """
        Store the time-step in the buffer.

        :param state: the current state to store in the buffer
        :param action: the last action to store in the buffer
        :param reward: the reward obtained from the action at the current state to store in the buffer
        :param value: the value of the state as estimated by the value stream of the _model to store in the buffer
        """
        # Append all data and increase the pointer
        self._states.append(state)
        self._actions.append(action)
        self._rewards.append(reward)
        self._values.append(value)
        self._pointer += 1

    def get(self) -> []:
        """
        Get all of the data from the buffer, with advantages appropriately normalized (shifted to have mean zero and
        standard deviation equals to one). Also reset pointers in the buffer and the lists composing the buffer.

        :return a list containing the ndarrays of: states, actions, advantages, rewards-to-go
        """
        # Get a numpy array on the advantage list
        advantages_array: numpy.ndarray = numpy.array(self._advantages)
        # Execute the advantage normalization trick
        global_sum: float = numpy.sum(advantages_array)
        advantage_mean: float = global_sum / advantages_array.size
        global_sum_squared: float = numpy.sum((advantages_array - advantage_mean) ** 2)
        advantage_std: float = numpy.sqrt(global_sum_squared / advantages_array.size)
        # Adjust advantages according to the trick
        advantages_array = ((advantages_array - advantage_mean) / advantage_std)
        # Save the necessary values as numpy arrays before reset
        states_array: numpy.ndarray = numpy.array(self._states)
        actions_array: numpy.ndarray = numpy.array(self._actions)
        rewards_to_go_array: numpy.ndarray = numpy.array(self._rewards_to_go)
        # Reset the buffer and related pointers
        self._pointer = 0
        self._path_start_index = 0
        self._states = []
        self._actions = []
        self._advantages = []
        self._rewards = []
        self._rewards_to_go = []
        self._values = []
        # Return all the buffer components
        return [states_array, actions_array, advantages_array, rewards_to_go_array]

    def finish_path(self,
                    value: float = 0):
        """
        Finish the path at the end of a trajectory. This looks back in the buffer to where the trajectory started,
        and uses rewards and value estimates from the whole trajectory to compute advantage estimates with GAE-Lambda,
        as well as compute the rewards-to-go for each state, to use as the targets for the value stream optimization.

        :param value: the last reward given by the environment or the last predicted value if last state is not terminal
        """
        path_slice = slice(self._path_start_index, self._pointer)
        rewards: numpy.ndarray = numpy.array(self._rewards[path_slice] + [value])
        values: numpy.ndarray = numpy.array(self._values[path_slice] + [value])
        # Compute GAE-Lambda advantage estimation (compute advantages using the value in the buffer taken from the model)
        deltas: numpy.ndarray = rewards[:-1] + self._discount_factor * values[1:] - values[:-1]
        self._advantages[path_slice] = self._discount_cumulative_sum(deltas, self._discount_factor * self._lambda_parameter).tolist()
        # Compute rewards-to-go
        self._rewards_to_go[path_slice] = (self._discount_cumulative_sum(rewards, self._discount_factor)[:-1]).tolist()
        self._path_start_index = self._pointer

    @staticmethod
    def _discount_cumulative_sum(vector: numpy.ndarray, discount: float) -> numpy.ndarray:
        """
        Compute discounted cumulative sums of vectors.
        Credits to rllab.

        :param vector: the vector on which to compute cumulative discounted sum (e.g. [x0, x1, x2])
        :return the discounted cumulative sum (e.g. [x0 + discount * x1 + discount^2 * x2, x1 + discount * x2, x3])
        """
        return scipy.signal.lfilter([1], [1, float(-discount)], vector[::-1], axis=0)[::-1]


class VanillaPolicyGradient(Model):
    """
    Vanilla Policy Gradient with GAE (Generalized Advantage Estimation).
    The algorithm is on-policy and executes updates every a certain number of episodes.
    The model is constituted by two sub-models, or streams. The first stream computes and optimizes the policy loss,
    and to drive the loss the advantages for each state in the batch is required to be estimated. This stream is called
    policy stream.
    The second stream computes the value on the current states and optimizes such estimation. To drive the loss of such
    sub-model, the value estimated by the stream itself for each state in the batch is required to be estimated. This
    stream is called value stream.
    The advantage used to drive the policy stream loss is computed using GAE on the value estimated by the value stream
    and such computation is carried on in the buffer.

    The buffer stores all the trajectories up to the update point. Since each episode can contains different
    numbers of steps, the buffer is dynamically resizable.
    The algorithm is very likely to converge to local minima.

    Supported observation spaces:
        - discrete
        - continuous

    Supported action spaces:
        - discrete
        - continuous
    """

    def __init__(self,
                 name: str,
                 discount_factor: float,
                 learning_rate_policy: float, learning_rate_value: float,
                 value_steps_for_update: int,
                 hidden_layers_config: Config,
                 lambda_parameter: float):
        # Define _model attributes
        self.learning_rate_policy: float = learning_rate_policy
        self.learning_rate_value: float = learning_rate_value
        self.discount_factor: float = discount_factor
        # Define internal _model attributes
        self._hidden_layers_config: Config = hidden_layers_config
        self._value_steps_for_update: int = value_steps_for_update
        self._lambda_parameter: float = lambda_parameter
        # Define vanilla policy gradient empty attributes
        self.buffer: Buffer = None
        # Define internal vanilla policy gradient empty attributes
        self._inputs = None
        self._actions = None
        self._targets = None
        self._advantages = None
        self._rewards = None
        self._mask = None
        self._logits = None
        self._expected_value = None
        self._std = None
        self._log_likelihood = None
        self._value = None
        self._value_stream_loss = None
        self._policy_stream_loss = None
        self._value_stream_optimizer = None
        self._policy_stream_optimizer = None
        # Generate the base _model
        super(VanillaPolicyGradient, self).__init__(name)
        # Define the types of allowed observation and action spaces
        self._supported_observation_space_types.append(SpaceType.discrete)
        self._supported_observation_space_types.append(SpaceType.continuous)
        self._supported_action_space_types.append(SpaceType.discrete)
        self._supported_action_space_types.append(SpaceType.continuous)

    def _define_graph(self):
        # Set the GAE buffer for the vanilla policy gradient algorithm
        self.buffer: Buffer = Buffer(self.discount_factor, self._lambda_parameter)
        # Define the tensorflow model
        with tensorflow.variable_scope(self._scope + "/" + self._name):
            # Define inputs of the estimator as a float adaptable array with shape Nx(S) where N is the number of examples and (S) the shape of the state
            self._inputs = tensorflow.placeholder(shape=[None, *self._observation_space_shape], dtype=tensorflow.float32, name="inputs")
            # Define the estimator network hidden layers from the config
            hidden_layers_output = self._hidden_layers_config.apply_hidden_layers(self._inputs)
            # Define the targets for learning with the same NxA adaptable size
            self._targets = tensorflow.placeholder(shape=(None, *self._agent_action_space_shape), dtype=tensorflow.float32, name="targets")
            # Change the _model definition according to its action space type
            if self._agent_action_space_type == SpaceType.discrete:
                # Define the mask placeholder
                self._mask = tensorflow.placeholder(shape=(None, *self._agent_action_space_shape), dtype=tensorflow.float32, name="mask")
                # Define the logits as outputs of the deep neural network with shape NxA where N is the number of inputs, A is the action size when its type is discrete
                self._logits = tensorflow.layers.dense(hidden_layers_output, *self._agent_action_space_shape, name="logits")
                # Compute the masked logits using the given mask
                masked_logits = tensorflow.add(self._logits, self._mask)
                # Define the actions on the first shape dimension as a squeeze on the samples drawn from a categorical distribution on the logits
                self._actions = tensorflow.squeeze(tensorflow.multinomial(logits=masked_logits, num_samples=1), axis=1)
                # Define the log likelihood according to the categorical distribution
                self._log_likelihood, _ = self.get_categorical_log_likelihood(self._targets, self._logits)
            else:
                # Define the expected value as the output of the deep neural network with shape Nx(A) where N is the number of inputs, (A) is the action shape
                self._expected_value = tensorflow.layers.dense(hidden_layers_output, *self._agent_action_space_shape, name="expected_value")
                # Define the log standard deviation
                log_std = tensorflow.get_variable(name="log_std", initializer=-0.5*numpy.ones(*self._agent_action_space_shape, dtype=numpy.float32))
                # Define the standard deviation
                self._std = tensorflow.exp(log_std, name="std")
                # Define actions as the expected value summed up with a noise vector multiplied by the standard deviation
                self._actions = self._expected_value + tensorflow.random_normal(tensorflow.shape(self._expected_value)) * self._std
                # Define the log likelihood according to the gaussian distribution
                self._log_likelihood = self.get_gaussian_log_likelihood(self._targets, self._expected_value, log_std)
            # Define the value estimator (a deep MLP)
            value_stream_hidden_layers_output = self._hidden_layers_config.apply_hidden_layers(self._inputs)
            value_stream_output = tensorflow.layers.dense(value_stream_hidden_layers_output, 1, activation=None)
            # Define value by squeezing the output of the advantage stream MLP
            self._value = tensorflow.squeeze(value_stream_output, axis=1, name="value")
            # Define the rewards as an adaptable vector of floats (they are actually rewards-to-go computed with GAE)
            self._rewards = tensorflow.placeholder(shape=(None, ), dtype=tensorflow.float32, name="rewards")
            # Define advantage loss as the mean squared error of the difference between computed rewards to go and the advantage
            self._value_stream_loss = tensorflow.reduce_mean((self._rewards - self._value) ** 2, name="value_loss")
            # Define the optimizer for the value stream (actually the MLP optimizer)
            self._value_stream_optimizer = tensorflow.train.AdamOptimizer(self.learning_rate_value).minimize(self._value_stream_loss)
            # Define the advantages placeholder as an adaptable vector of floats (they are computed by the MLP and stored in the buffer)
            # Note: the model get the advantages from the buffer once computed using GAE on the values
            self._advantages = tensorflow.placeholder(shape=(None,), dtype=tensorflow.float32, name="advantages")
            # Define the loss as the mean of the rewards multiplied the log likelihood
            self._policy_stream_loss = -tensorflow.reduce_mean(self._advantages * self._log_likelihood, name="policy_loss")
            # Define the optimizer for the policy stream
            self._policy_stream_optimizer = tensorflow.train.AdamOptimizer(self.learning_rate_policy).minimize(self._policy_stream_loss)
            # Define the _initializer
            self._initializer = tensorflow.global_variables_initializer()

    def _define_summary(self):
        with tensorflow.variable_scope(self._scope + "/" + self._name):
            # Define the _summary operation for this graph with losses summaries
            self.summary = tensorflow.summary.merge([tensorflow.summary.scalar("policy_stream_loss", self._policy_stream_loss),
                                                     tensorflow.summary.scalar("value_stream_loss", self._value_stream_loss)])

    def sample_action(self,
                      session,
                      observation_current,
                      mask: numpy.ndarray = None):
        """
        Get the action sampled from the probability distribution of the model given the current observation and an optional mask.

        :param session: the session of tensorflow currently running
        :param observation_current: the current observation of the agent in the environment to base prediction upon
        :param mask: the optional mask used (only if the action space type is discrete) to remove certain actions from the prediction (-infinity to remove, 0.0 to pass-through)
        :return: the action predicted by the model
        """
        # If there is no mask and the action space type is discrete generate a full pass-through mask
        if mask is None and self._agent_action_space_type == SpaceType.discrete:
            mask = numpy.zeros(self._agent_action_space_shape, dtype=float)
        # Return a random action sample given the current state and depending on the observation space type and also compute value estimate
        if self._observation_space_type == SpaceType.discrete:
            # Generate a one-hot encoded version of the observation if observation space type is discrete
            observation_current_one_hot: numpy.ndarray = numpy.identity(*self._observation_space_shape)[observation_current]
            if self._agent_action_space_type == SpaceType.discrete:
                action, value = session.run([self._actions, self._value],
                                            feed_dict={self._inputs: [observation_current_one_hot], self._mask: [mask]})
            else:
                action, value = session.run([self._actions, self._value],
                                            feed_dict={self._inputs: [observation_current_one_hot]})
        else:
            if self._agent_action_space_type == SpaceType.discrete:
                action, value = session.run([self._actions, self._value],
                                            feed_dict={self._inputs: [observation_current], self._mask: [mask]})
            else:
                action, value = session.run([self._actions, self._value],
                                            feed_dict={self._inputs: [observation_current]})
        # Return the predicted action and the estimated value
        return action[0], value[0]

    def get_action_probabilities(self,
                                 session,
                                 observation_current,
                                 mask: numpy.ndarray = None) -> []:
        """
        Get all the action probabilities (logits if discrete, expected value if continuous) for the given current observation and
        an optional mask.

        :param session: the session of tensorflow currently running
        :param observation_current: the current observation of the agent in the environment to base prediction upon
        :param mask: the optional mask used (only if the action space type is discrete) to remove certain actions from the prediction (-infinity to remove, 0.0 to pass-through)
        :return: the list of action probabilities (logits or expected values depending on the agent action space type)
        """
        if mask is None and self._agent_action_space_type == SpaceType.discrete:
            mask = numpy.zeros(self._agent_action_space_shape, dtype=float)
        # Get the logits or the expected value as the distribution of the action probabilities depending on the action space shape
        if self._agent_action_space_type == SpaceType.discrete:
            if self._observation_space_type == SpaceType.discrete:
                observation_current_one_hot: numpy.ndarray = numpy.identity(*self._observation_space_shape)[observation_current]
                if self._agent_action_space_type == SpaceType.discrete:
                    logits = session.run([self._logits],
                                         feed_dict={self._inputs: [observation_current_one_hot], self._mask: [mask]})
                else:
                    logits = session.run([self._logits],
                                         feed_dict={self._inputs: [observation_current_one_hot]})
            else:
                if self._agent_action_space_type == SpaceType.discrete:
                    logits = session.run([self._logits],
                                         feed_dict={self._inputs: [observation_current]})
                else:
                    logits = session.run([self._logits],
                                         feed_dict={self._inputs: [observation_current], self._mask: [mask]})
            # Return the logits (probabilities of all actions)
            return logits
        else:
            expected_value, std = session.run([self._expected_value, self._std], feed_dict={self._inputs: [observation_current]})
            # Return the expected value and std of all actions
            return expected_value

    def update(self,
               session,
               batch: []):
        # Unpack the batch
        observations, actions, advantages, rewards = batch[0], batch[1], batch[2], batch[3]
        # Generate a one-hot encoded version of the observations if observation space type is discrete
        if self._observation_space_type == SpaceType.discrete:
            observations: numpy.ndarray = numpy.eye(*self._observation_space_shape)[numpy.array(observations).reshape(-1)]
        # Generate a one-hot encoded version of the actions if action space type is discrete
        if self._agent_action_space_type == SpaceType.discrete:
            actions: numpy.ndarray = numpy.eye(*self._agent_action_space_shape)[numpy.array(actions).reshape(-1)]
        # Run the policy optimizer of the _model in training mode
        session.run(self._policy_stream_optimizer,
                    feed_dict={
                                self._inputs: observations,
                                self._targets: actions,
                                self._advantages: advantages
                              })
        # Run the value optimizer of the _model in training mode for the required amount of steps
        for _ in range(self._value_steps_for_update):
            session.run(self._value_stream_optimizer,
                        feed_dict={
                                    self._inputs: observations,
                                    self._advantages: advantages,
                                    self._rewards: rewards
                                  })
        # Compute the policy loss and value loss of the _model after this sequence of training and also compute the _summary
        policy_loss, value_loss, summary = session.run([self._policy_stream_loss, self._value_stream_loss, self.summary],
                                                       feed_dict={
                                                                   self._inputs: observations,
                                                                   self._targets: actions,
                                                                   self._rewards: rewards,
                                                                   self._advantages: advantages
                                                                 })
        # Return both losses and _summary for the update sequence
        return summary, policy_loss, value_loss

    @property
    def warmup_episodes(self) -> int:
        return 0

    @staticmethod
    def get_categorical_log_likelihood(target_actions_mask, logits):
        """
        Get log-likelihood for discrete action spaces (using a categorical distribution) on the given logits and with
        the target action mask.
        It uses tensorflow and as such should only be called in the _define method.

        :param target_actions_mask: the target actions used to mask the log-likelihood on the logits
        :param logits: the logits of the neural network
        :return: the log-likelihood according to categorical distribution
        """
        # Define the unmasked likelihood as the log-softmax of the logits
        log_likelihood_unmasked = tensorflow.nn.log_softmax(logits)
        # Return the categorical log-likelihood by summing over the first axis of the target action mask multiplied
        # by the log-likelihood on the logits (unmasked, this is the masking operation) and the unmasked likelihood
        return tensorflow.reduce_sum(target_actions_mask * log_likelihood_unmasked, axis=1, name="log_likelihood"), log_likelihood_unmasked

    @staticmethod
    def get_gaussian_log_likelihood(target_actions, expected_value, log_std):
        """
        Get log-likelihood for continuous action spaces (using a gaussian distribution) on the given expected value and
        log-std and with the target actions.
        It uses tensorflow and as such should only be called in the _define method.

        :param target_actions: the target actions used to compute the log-likelihood tensor
        :param expected_value: the expected value of the gaussian distribution
        :param log_std: the log-std of the gaussian distribution
        :return: the log-likelihood according to gaussian distribution
        """
        # Define the log-likelihood tensor for the gaussian distribution on the given target actions
        log_likelihood_tensor = -0.5 * (((target_actions - expected_value) / (tensorflow.exp(log_std) + 1e-8)) ** 2 + 2 * log_std + numpy.log(2 * numpy.pi))
        # Return the gaussian log-likelihood by summing over all the elements in the log-likelihood tensor defined above
        return tensorflow.reduce_sum(log_likelihood_tensor, axis=1, name="log_likelihood")
